#!/usr/bin/env python3
"""Tests for TextProcessor"""
import sys
import unittest

from gruut.text_processor import Sentence, TextProcessor, TextProcessorSettings, Word
from gruut.utils import print_graph

WORDS_KWARGS = {"explicit_lang": False, "phonemes": False, "pos": False}


class TextProcessorTestCase(unittest.TestCase):
    """Tests for TextProcessor"""

    def test_whitespace(self):
        """Text whitespace preservation"""
        processor = TextProcessor()
        graph, root = processor("This is  a   test    ")
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # Whitespace is retained by default
        self.assertEqual(
            words,
            [
                Word(idx=0, sent_idx=0, text="This", text_with_ws="This "),
                Word(idx=1, sent_idx=0, text="is", text_with_ws="is  "),
                Word(idx=2, sent_idx=0, text="a", text_with_ws="a   "),
                Word(idx=3, sent_idx=0, text="test", text_with_ws="test    "),
            ],
        )

    def test_no_whitespace(self):
        """Test disabling of whitespace preservation"""
        processor = TextProcessor(keep_whitespace=False)
        graph, root = processor("This is  a   test    ")
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # Whitespace is discarded
        self.assertEqual(
            words,
            [
                Word(idx=0, sent_idx=0, text="This", text_with_ws="This"),
                Word(idx=1, sent_idx=0, text="is", text_with_ws="is"),
                Word(idx=2, sent_idx=0, text="a", text_with_ws="a"),
                Word(idx=3, sent_idx=0, text="test", text_with_ws="test"),
            ],
        )

    def test_punctuation(self):
        """Test splitting of punctuation from around words"""
        processor = TextProcessor(
            begin_punctuations={'"', "«"},
            end_punctuations={'"', "»"},
            minor_breaks={","},
            major_breaks={"."},
        )
        graph, root = processor('This «is»,  a "test".')
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # Punctuations are separated
        self.assertEqual(
            words,
            [
                Word(idx=0, sent_idx=0, text="This", text_with_ws="This "),
                Word(
                    idx=1, sent_idx=0, text="«", text_with_ws="«", is_punctuation=True
                ),
                Word(idx=2, sent_idx=0, text="is", text_with_ws="is"),
                Word(
                    idx=3, sent_idx=0, text="»", text_with_ws="»", is_punctuation=True
                ),
                Word(
                    idx=4, sent_idx=0, text=",", text_with_ws=",  ", is_minor_break=True
                ),
                Word(idx=5, sent_idx=0, text="a", text_with_ws="a "),
                Word(
                    idx=6, sent_idx=0, text='"', text_with_ws='"', is_punctuation=True
                ),
                Word(idx=7, sent_idx=0, text="test", text_with_ws="test"),
                Word(
                    idx=8, sent_idx=0, text='"', text_with_ws='"', is_punctuation=True
                ),
                Word(
                    idx=9, sent_idx=0, text=".", text_with_ws=".", is_major_break=True
                ),
            ],
        )

    def test_punctuation_with_inner_break(self):
        """Test break inside of punctuation"""
        processor = TextProcessor(
            begin_punctuations={'"'}, end_punctuations={'"'}, major_breaks={"."},
        )
        graph, root = processor('Test "one." Test two.')
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # First sentence includes final quote
        self.assertEqual(
            words,
            [
                # First sentence
                Word(idx=0, sent_idx=0, text="Test", text_with_ws="Test "),
                Word(
                    idx=1, sent_idx=0, text='"', text_with_ws='"', is_punctuation=True
                ),
                Word(idx=2, sent_idx=0, text="one", text_with_ws="one"),
                Word(
                    idx=3, sent_idx=0, text=".", text_with_ws=".", is_major_break=True
                ),
                Word(
                    idx=4, sent_idx=0, text='"', text_with_ws='" ', is_punctuation=True
                ),
                # Second sentence
                Word(idx=0, sent_idx=1, text="Test", text_with_ws="Test "),
                Word(idx=1, sent_idx=1, text="two", text_with_ws="two"),
                Word(
                    idx=2, sent_idx=1, text=".", text_with_ws=".", is_major_break=True
                ),
            ],
        )

    def test_replacements(self):
        """Test regex replacements during tokenization"""
        processor = TextProcessor(
            minor_breaks={","},
            major_breaks={"."},
            replacements=[
                ("\\B'", '"'),  # replace single quotes
                ("'\\B", '"'),
                ('[\\<\\>\\(\\)\\[\\]"]+', ""),  # drop brackets/quotes
            ],
        )
        graph, root = processor("\"This,\" [is] <a> (test) 'sentence.'")
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # Quotes and brackets are discarded
        self.assertEqual(
            words,
            [
                Word(idx=0, sent_idx=0, text="This", text_with_ws="This"),
                Word(
                    idx=1, sent_idx=0, text=",", text_with_ws=", ", is_minor_break=True
                ),
                Word(idx=2, sent_idx=0, text="is", text_with_ws="is "),
                Word(idx=3, sent_idx=0, text="a", text_with_ws="a "),
                Word(idx=4, sent_idx=0, text="test", text_with_ws="test "),
                Word(idx=5, sent_idx=0, text="sentence", text_with_ws="sentence"),
                Word(
                    idx=6, sent_idx=0, text=".", text_with_ws=".", is_major_break=True
                ),
            ],
        )

    def test_abbreviations(self):
        """Test expansion of abbreviations (with case preservation)"""
        processor = TextProcessor(
            minor_breaks={","},
            major_breaks={".", "?"},
            abbreviations={
                r"^([dD])r\.": r"\1octor",
                r"^([mM])r\.": r"\1ister",
                r"^([sS])t\.": r"\1treet",
            },
        )
        graph, root = processor("Mr.? I'm just a dr., on this St. at least.")
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # Abbreviations are expanded, maintaining capitalization
        self.assertEqual(
            words,
            [
                Word(idx=0, sent_idx=0, text="Mister", text_with_ws="Mister"),
                Word(
                    idx=1, sent_idx=0, text="?", text_with_ws="? ", is_major_break=True
                ),
                Word(idx=0, sent_idx=1, text="I'm", text_with_ws="I'm "),
                Word(idx=1, sent_idx=1, text="just", text_with_ws="just "),
                Word(idx=2, sent_idx=1, text="a", text_with_ws="a "),
                Word(idx=3, sent_idx=1, text="doctor", text_with_ws="doctor"),
                Word(
                    idx=4, sent_idx=1, text=",", text_with_ws=", ", is_minor_break=True
                ),
                Word(idx=5, sent_idx=1, text="on", text_with_ws="on "),
                Word(idx=6, sent_idx=1, text="this", text_with_ws="this "),
                Word(idx=7, sent_idx=1, text="Street", text_with_ws="Street "),
                Word(idx=8, sent_idx=1, text="at", text_with_ws="at "),
                Word(idx=9, sent_idx=1, text="least", text_with_ws="least"),
                Word(
                    idx=10, sent_idx=1, text=".", text_with_ws=".", is_major_break=True
                ),
            ],
        )

    def test_multiple_sentences(self):
        """Test sentence break"""
        processor = TextProcessor(major_breaks={".", "!"})
        graph, root = processor("First  sentence. Second sentence! ")
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # Separated by a major break
        self.assertEqual(
            words,
            [
                Word(idx=0, sent_idx=0, text="First", text_with_ws="First  "),
                Word(idx=1, sent_idx=0, text="sentence", text_with_ws="sentence"),
                Word(
                    idx=2, sent_idx=0, text=".", text_with_ws=". ", is_major_break=True
                ),
                Word(idx=0, sent_idx=1, text="Second", text_with_ws="Second "),
                Word(idx=1, sent_idx=1, text="sentence", text_with_ws="sentence"),
                Word(
                    idx=2, sent_idx=1, text="!", text_with_ws="! ", is_major_break=True
                ),
            ],
        )

        # Check sentences too
        sentences = list(processor.sentences(graph, root, **WORDS_KWARGS))
        self.assertEqual(
            sentences,
            [
                Sentence(
                    idx=0,
                    text="First sentence.",
                    text_with_ws="First  sentence. ",
                    text_spoken="First sentence",
                    words=[
                        Word(idx=0, sent_idx=0, text="First", text_with_ws="First  "),
                        Word(
                            idx=1, sent_idx=0, text="sentence", text_with_ws="sentence"
                        ),
                        Word(
                            idx=2,
                            sent_idx=0,
                            text=".",
                            text_with_ws=". ",
                            is_major_break=True,
                        ),
                    ],
                ),
                Sentence(
                    idx=1,
                    text="Second sentence!",
                    text_with_ws="Second sentence! ",
                    text_spoken="Second sentence",
                    words=[
                        Word(idx=0, sent_idx=1, text="Second", text_with_ws="Second "),
                        Word(
                            idx=1, sent_idx=1, text="sentence", text_with_ws="sentence"
                        ),
                        Word(
                            idx=2,
                            sent_idx=1,
                            text="!",
                            text_with_ws="! ",
                            is_major_break=True,
                        ),
                    ],
                ),
            ],
        )

    def test_multiple_paragraphs(self):
        """Test paragraph index"""
        processor = TextProcessor()
        graph, root = processor(
            "<speak><p>First paragraph</p><p>Second paragraph</p></speak>", ssml=True
        )
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # Sentences/words should be in different paragraphs
        self.assertEqual(
            words,
            [
                Word(idx=0, sent_idx=0, par_idx=0, text="First", text_with_ws="First "),
                Word(
                    idx=1,
                    sent_idx=0,
                    par_idx=0,
                    text="paragraph",
                    text_with_ws="paragraph",
                ),
                Word(
                    idx=0, sent_idx=0, par_idx=1, text="Second", text_with_ws="Second "
                ),
                Word(
                    idx=1,
                    sent_idx=0,
                    par_idx=1,
                    text="paragraph",
                    text_with_ws="paragraph",
                ),
            ],
        )

    def test_explicit_sentence(self):
        """Test <s> in SSML for avoiding sentence break"""
        processor = TextProcessor(major_breaks={".", "!"})
        graph, root = processor("<s>First sentence. Second sentence!</s>", ssml=True)
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # Sentences should not be split apart
        self.assertEqual(
            words,
            [
                Word(idx=0, sent_idx=0, text="First", text_with_ws="First "),
                Word(idx=1, sent_idx=0, text="sentence", text_with_ws="sentence"),
                Word(
                    idx=2, sent_idx=0, text=".", text_with_ws=". ", is_major_break=True
                ),
                Word(idx=3, sent_idx=0, text="Second", text_with_ws="Second "),
                Word(idx=4, sent_idx=0, text="sentence", text_with_ws="sentence"),
                Word(
                    idx=5, sent_idx=0, text="!", text_with_ws="!", is_major_break=True
                ),
            ],
        )

    def test_minor_breaks(self):
        """Test minor (phrase) break"""
        processor = TextProcessor(minor_breaks={","})
        graph, root = processor("this, is a test")
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # Comma should be split from word
        self.assertEqual(
            words,
            [
                Word(idx=0, sent_idx=0, text="this", text_with_ws="this"),
                Word(
                    idx=1, sent_idx=0, text=",", text_with_ws=", ", is_minor_break=True
                ),
                Word(idx=2, sent_idx=0, text="is", text_with_ws="is "),
                Word(idx=3, sent_idx=0, text="a", text_with_ws="a "),
                Word(idx=4, sent_idx=0, text="test", text_with_ws="test"),
            ],
        )

    def test_word_breaks(self):
        """Test inner-word break"""
        processor = TextProcessor(word_breaks={"-"})
        graph, root = processor("ninety-nine")
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # Word should be split
        self.assertEqual(
            words,
            [
                Word(idx=0, sent_idx=0, text="ninety", text_with_ws="ninety "),
                Word(idx=1, sent_idx=0, text="nine", text_with_ws="nine"),
            ],
        )

    def test_spell_out(self):
        """Test interpret-as="spell-out" in SSML"""
        processor = TextProcessor(default_lang="en_US")
        graph, root = processor(
            '<say-as interpret-as="spell-out">test123</say-as>', ssml=True
        )
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        print_graph(graph, root)

        # Word should be split into letters
        self.assertEqual(
            words,
            [
                Word(idx=0, sent_idx=0, text="t", text_with_ws="t "),
                Word(idx=1, sent_idx=0, text="e", text_with_ws="e "),
                Word(idx=2, sent_idx=0, text="s", text_with_ws="s "),
                Word(idx=3, sent_idx=0, text="t", text_with_ws="t "),
                Word(idx=4, sent_idx=0, text="one", text_with_ws="one "),
                Word(idx=5, sent_idx=0, text="two", text_with_ws="two "),
                Word(idx=6, sent_idx=0, text="three", text_with_ws="three"),
            ],
        )

    def test_initialisms(self):
        """Test initialism spell out"""
        processor = TextProcessor(
            major_breaks={"."},
            is_initialism=lambda s: s.isalpha() and s.isupper(),
            split_initialism=list,
        )
        graph, root = processor("TTS.")
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # Letters should be split
        self.assertEqual(
            words,
            [
                Word(idx=0, sent_idx=0, text="T", text_with_ws="T "),
                Word(idx=1, sent_idx=0, text="T", text_with_ws="T "),
                Word(idx=2, sent_idx=0, text="S", text_with_ws="S"),
                Word(
                    idx=3, sent_idx=0, text=".", text_with_ws=".", is_major_break=True
                ),
            ],
        )

    def test_numbers_one_language(self):
        """Test number verbalization (single language)"""
        processor = TextProcessor(default_lang="en_US")
        graph, root = processor("1 2 3")
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # Numbers should be verbalized
        self.assertEqual(
            words,
            [
                Word(idx=0, sent_idx=0, text="one", text_with_ws="one "),
                Word(idx=1, sent_idx=0, text="two", text_with_ws="two "),
                Word(idx=2, sent_idx=0, text="three", text_with_ws="three"),
            ],
        )

    def test_numbers_multiple_languages(self):
        """Test number verbalization (SSML, multiple languages)"""
        processor = TextProcessor(default_lang="en_US")
        graph, root = processor(
            '1 <w lang="es_ES">2</w> <w lang="de_DE">3</w>', ssml=True
        )
        words = list(processor.words(graph, root, phonemes=False))

        # Numbers should be verbalized
        self.assertEqual(
            words,
            [
                Word(lang="en_US", idx=0, sent_idx=0, text="one", text_with_ws="one "),
                Word(lang="es_ES", idx=1, sent_idx=0, text="dos", text_with_ws="dos "),
                Word(lang="de_DE", idx=2, sent_idx=0, text="drei", text_with_ws="drei"),
            ],
        )

    def test_currency_one_language(self):
        """Test currency verbalization (single language)"""
        processor = TextProcessor(default_lang="en_US")
        graph, root = processor("$10")
        words = list(processor.words(graph, root, phonemes=False, pos=False))

        # Currency should be verbalized
        self.assertEqual(
            words,
            [
                Word(lang="en_US", idx=0, sent_idx=0, text="ten", text_with_ws="ten "),
                Word(
                    lang="en_US",
                    idx=1,
                    sent_idx=0,
                    text="dollars",
                    text_with_ws="dollars",
                ),
            ],
        )

    def test_currency_multiple_language(self):
        """Test currency verbalization (SSML, multiple languages)"""
        processor = TextProcessor(default_lang="en_US")
        graph, root = processor(
            '€10 <w lang="fr_FR">€10</w> <w lang="nl_NL">€10</w>',
            ssml=True,
            phonemize=False,
        )
        words = list(processor.words(graph, root, phonemes=False, pos=False))

        # Currencies should be verbalized
        self.assertEqual(
            words,
            [
                Word(lang="en_US", idx=0, sent_idx=0, text="ten", text_with_ws="ten "),
                Word(
                    lang="en_US", idx=1, sent_idx=0, text="euro", text_with_ws="euro "
                ),
                Word(lang="fr_FR", idx=2, sent_idx=0, text="dix", text_with_ws="dix "),
                Word(
                    lang="fr_FR", idx=3, sent_idx=0, text="euros", text_with_ws="euros "
                ),
                Word(
                    lang="nl_NL", idx=4, sent_idx=0, text="tien", text_with_ws="tien "
                ),
                Word(lang="nl_NL", idx=5, sent_idx=0, text="euro", text_with_ws="euro"),
            ],
        )

    def test_currency_default(self):
        """Test default currency use when no currency symbol (interpret-as="currency")"""
        processor = TextProcessor(default_lang="en_US", default_currency="USD")
        graph, root = processor(
            '<say-as interpret-as="currency">10</say-as>', ssml=True
        )
        words = list(processor.words(graph, root, phonemes=False, pos=False))

        # Currency should be verbalized, despite lack of "$" symbol
        self.assertEqual(
            words,
            [
                Word(lang="en_US", idx=0, sent_idx=0, text="ten", text_with_ws="ten "),
                Word(
                    lang="en_US",
                    idx=1,
                    sent_idx=0,
                    text="dollars",
                    text_with_ws="dollars",
                ),
            ],
        )

    def test_time(self):
        """Test time verbalization (English)"""
        processor = TextProcessor(default_lang="en_US")
        graph, root = processor("  4:01pm")
        words = list(processor.words(graph, root, phonemes=False, pos=False))

        # Time should be verbalized
        self.assertEqual(
            words,
            [
                Word(
                    lang="en_US", idx=0, sent_idx=0, text="four", text_with_ws="  four "
                ),
                Word(lang="en_US", idx=1, sent_idx=0, text="oh", text_with_ws="oh "),
                Word(lang="en_US", idx=2, sent_idx=0, text="one", text_with_ws="one "),
                Word(lang="en_US", idx=3, sent_idx=0, text="P", text_with_ws="P "),
                Word(lang="en_US", idx=4, sent_idx=0, text="M", text_with_ws="M"),
            ],
        )

    def test_time_no_colon(self):
        """Test time verbalization without a colon (English)"""
        processor = TextProcessor(default_lang="en_US")
        graph, root = processor("10am")
        words = list(processor.words(graph, root, phonemes=False, pos=False))

        # Time should be verbalized
        self.assertEqual(
            words,
            [
                Word(lang="en_US", idx=0, sent_idx=0, text="ten", text_with_ws="ten "),
                Word(lang="en_US", idx=1, sent_idx=0, text="A", text_with_ws="A "),
                Word(lang="en_US", idx=2, sent_idx=0, text="M", text_with_ws="M"),
            ],
        )

    def test_date_one_language(self):
        """Test date verbalization (single language)"""
        processor = TextProcessor(default_lang="en_US", word_breaks={"-"})
        graph, root = processor("4/1/1999")
        words = list(processor.words(graph, root, phonemes=False, pos=False))

        # Date should be verbalized
        self.assertEqual(
            words,
            [
                Word(
                    lang="en_US", idx=0, sent_idx=0, text="April", text_with_ws="April "
                ),
                Word(
                    lang="en_US", idx=1, sent_idx=0, text="first", text_with_ws="first"
                ),
                Word(
                    lang="en_US",
                    idx=2,
                    sent_idx=0,
                    text=",",
                    text_with_ws=", ",
                    is_minor_break=True,
                ),
                Word(
                    lang="en_US",
                    idx=3,
                    sent_idx=0,
                    text="nineteen",
                    text_with_ws="nineteen ",
                ),
                Word(
                    lang="en_US",
                    idx=4,
                    sent_idx=0,
                    text="ninety",
                    text_with_ws="ninety ",
                ),
                Word(lang="en_US", idx=5, sent_idx=0, text="nine", text_with_ws="nine"),
            ],
        )

    def test_date_multiple_languages(self):
        """Test date verbalization (SSML, multiple languages)"""
        processor = TextProcessor(default_lang="en_US", word_breaks={"-"})
        graph, root = processor(
            '<speak><s>4/1/1999</s> <s lang="fr_FR">4/1/1999</s><s lang="de_DE">01.04.1999</s></speak>',
            ssml=True,
            phonemize=False,  # ensure French year is split
        )
        words = list(processor.words(graph, root, phonemes=False, pos=False))

        # Date should be verbalized
        self.assertEqual(
            words,
            [
                # English
                Word(
                    lang="en_US", idx=0, sent_idx=0, text="April", text_with_ws="April "
                ),
                Word(
                    lang="en_US", idx=1, sent_idx=0, text="first", text_with_ws="first"
                ),
                Word(
                    lang="en_US",
                    idx=2,
                    sent_idx=0,
                    text=",",
                    text_with_ws=", ",
                    is_minor_break=True,
                ),
                Word(
                    lang="en_US",
                    idx=3,
                    sent_idx=0,
                    text="nineteen",
                    text_with_ws="nineteen ",
                ),
                Word(
                    lang="en_US",
                    idx=4,
                    sent_idx=0,
                    text="ninety",
                    text_with_ws="ninety ",
                ),
                Word(lang="en_US", idx=5, sent_idx=0, text="nine", text_with_ws="nine"),
                # French
                Word(
                    lang="fr_FR",
                    idx=0,
                    sent_idx=1,
                    text="quatrième",
                    text_with_ws="quatrième ",
                ),
                Word(
                    lang="fr_FR",
                    idx=1,
                    sent_idx=1,
                    text="janvier",
                    text_with_ws="janvier ",
                ),
                Word(
                    lang="fr_FR", idx=2, sent_idx=1, text="mille", text_with_ws="mille "
                ),
                Word(
                    lang="fr_FR", idx=3, sent_idx=1, text="neuf", text_with_ws="neuf "
                ),
                Word(
                    lang="fr_FR", idx=4, sent_idx=1, text="cent", text_with_ws="cent "
                ),
                Word(
                    lang="fr_FR",
                    idx=5,
                    sent_idx=1,
                    text="quatre",
                    text_with_ws="quatre ",
                ),
                Word(
                    lang="fr_FR", idx=6, sent_idx=1, text="vingt", text_with_ws="vingt "
                ),
                Word(lang="fr_FR", idx=7, sent_idx=1, text="dix", text_with_ws="dix "),
                Word(lang="fr_FR", idx=8, sent_idx=1, text="neuf", text_with_ws="neuf"),
                # German
                Word(
                    lang="de_DE",
                    idx=0,
                    sent_idx=2,
                    text="erste",
                    text_with_ws="erste ",
                ),
                Word(
                    lang="de_DE",
                    idx=1,
                    sent_idx=2,
                    text="April",
                    text_with_ws="April ",
                ),
                Word(
                    lang="de_DE",
                    idx=2,
                    sent_idx=2,
                    text="neunzehnhundertneunundneunzig",
                    text_with_ws="neunzehnhundertneunundneunzig",
                ),
            ],
        )

    def test_date_format_ordinal(self):
        """Test date format in SSML (ordinal)"""
        processor = TextProcessor(default_lang="en_US")
        graph, root = processor(
            '<say-as interpret-as="date" format="md">4/1</say-as>', ssml=True
        )
        words = list(processor.words(graph, root, phonemes=False, pos=False))

        # Date is forced to be interpreted and format using day ordinal (first)
        self.assertEqual(
            words,
            [
                Word(
                    lang="en_US", idx=0, sent_idx=0, text="April", text_with_ws="April "
                ),
                Word(lang="en_US", idx=1, sent_idx=0, text="one", text_with_ws="one"),
            ],
        )

    def test_date_format_cardinal(self):
        """Test date format in SSML (cardinal)"""
        processor = TextProcessor(default_lang="en_US")
        graph, root = processor(
            '<say-as interpret-as="date" format="dmy">4/1/2000</say-as>', ssml=True
        )
        words = list(processor.words(graph, root, phonemes=False, pos=False))

        # Date is forced to be interpreted and format using day ordinal (first)
        self.assertEqual(
            words,
            [
                Word(lang="en_US", idx=0, sent_idx=0, text="one", text_with_ws="one "),
                Word(
                    lang="en_US", idx=1, sent_idx=0, text="April", text_with_ws="April "
                ),
                Word(lang="en_US", idx=2, sent_idx=0, text="two", text_with_ws="two "),
                Word(
                    lang="en_US",
                    idx=3,
                    sent_idx=0,
                    text="thousand",
                    text_with_ws="thousand",
                ),
            ],
        )

    def test_part_of_speech_tagging(self):
        """Test part-of-speech tagging"""

        def get_parts_of_speech(words, *args, **kwargs):
            return [w.upper() for w in words]

        processor = TextProcessor(
            # Made-up tagger that just gives the UPPER of the word back
            get_parts_of_speech=get_parts_of_speech
        )
        graph, root = processor("a test")
        words = list(processor.words(graph, root, explicit_lang=False, phonemes=False))

        # Fake POS tags are added
        self.assertEqual(
            words,
            [
                Word(idx=0, sent_idx=0, text="a", text_with_ws="a ", pos="A"),
                Word(idx=1, sent_idx=0, text="test", text_with_ws="test", pos="TEST"),
            ],
        )

    def test_phonemize_one_language(self):
        """Test phonemizer (single language)"""

        def lookup_phonemes(word: str, *args, **kwargs):
            return list(word)

        processor = TextProcessor(
            # Made-up phonemizer that just gives back the letters
            lookup_phonemes=lookup_phonemes,
        )
        graph, root = processor("test")
        words = list(processor.words(graph, root, pos=False, explicit_lang=False))

        # Single word is "phonemized"
        self.assertEqual(
            words,
            [
                Word(
                    idx=0,
                    sent_idx=0,
                    text="test",
                    text_with_ws="test",
                    phonemes=["t", "e", "s", "t"],
                ),
            ],
        )

    def test_phonemize_one_language_multiple_roles(self):
        """Test phonemizer (SSML, multiple word roles)"""

        def lookup_phonemes(word, role=None, **kwargs):
            return list(word) if not role else list(word.upper())

        processor = TextProcessor(
            # Made-up phonemizer that gives back upper-case letters if a role is provided
            lookup_phonemes=lookup_phonemes
        )

        # Use made-up role
        graph, root = processor(
            '<speak>test <w role="some_role">test</w></speak>', ssml=True, pos=False
        )
        words = list(processor.words(graph, root, pos=False, explicit_lang=False))

        # Single word is phonemized two different manners depending on role
        self.assertEqual(
            words,
            [
                Word(
                    idx=0,
                    sent_idx=0,
                    text="test",
                    text_with_ws="test ",
                    phonemes=["t", "e", "s", "t"],
                ),
                Word(
                    idx=1,
                    sent_idx=0,
                    text="test",
                    text_with_ws="test",
                    phonemes=["T", "E", "S", "T"],
                ),
            ],
        )

    def test_phonemize_multiple_languages(self):
        """Test phonemizer (SSML, multiple languages)"""

        def en_lookup_phonemes(word: str, *args, **kwargs):
            return list(word)

        def de_lookup_phonemes(word: str, *args, **kwargs):
            return list(word.upper())

        processor = TextProcessor(
            default_lang="en_US",
            lookup_phonemes=en_lookup_phonemes,
            settings={
                "de_DE": TextProcessorSettings(
                    lang="de_DE", lookup_phonemes=de_lookup_phonemes
                )
            },
        )
        graph, root = processor(
            '<speak>test <w lang="de_DE">test</w></speak>', ssml=True
        )
        words = list(processor.words(graph, root))

        # Single word is phonemized according to the lexicon with two different languages
        self.assertEqual(
            words,
            [
                Word(
                    lang="en_US",
                    idx=0,
                    sent_idx=0,
                    text="test",
                    text_with_ws="test ",
                    phonemes=["t", "e", "s", "t"],
                ),
                Word(
                    lang="de_DE",
                    idx=1,
                    sent_idx=0,
                    text="test",
                    text_with_ws="test",
                    phonemes=["T", "E", "S", "T"],
                ),
            ],
        )

    def test_sub(self):
        """Test SSML substitution"""
        processor = TextProcessor(default_lang="en_US")
        graph, root = processor(
            '<speak><sub alias="World Wide Web Consortium">W3C</sub></speak>', ssml=True
        )
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # Single word is replaced by multiple words
        self.assertEqual(
            words,
            [
                Word(idx=0, sent_idx=0, text="World", text_with_ws="World ",),
                Word(idx=1, sent_idx=0, text="Wide", text_with_ws="Wide ",),
                Word(idx=2, sent_idx=0, text="Web", text_with_ws="Web ",),
                Word(idx=3, sent_idx=0, text="Consortium", text_with_ws="Consortium",),
            ],
        )

    def test_break(self):
        """Test SSML break tag"""
        processor = TextProcessor(default_lang="en_US", keep_whitespace=False)
        graph, root = processor(
            """
        <speak>
          <break time="1s"/>
          <p>
            <break time="2s" />
            <s>
              <break time="3s" />
              Break <break time="4s" /> here
            </s>
            <break time="5s" />
          </p>
          <break time="6s" />
        </speak>
        """,
            ssml=True,
        )
        sentences = list(processor.sentences(graph, root, **WORDS_KWARGS))

        # Break times are attached to appropriate elements
        self.assertEqual(
            sentences,
            [
                Sentence(
                    idx=0,
                    text="Break here",
                    text_with_ws="Break here",
                    text_spoken="Break here",
                    pause_before_ms=((1 + 2) * 1000),
                    pause_after_ms=((5 + 6) * 1000),
                    words=[
                        Word(
                            idx=0,
                            sent_idx=0,
                            text="Break",
                            text_with_ws="Break",
                            pause_before_ms=(3 * 1000),
                            pause_after_ms=(4 * 1000),
                        ),
                        Word(idx=1, sent_idx=0, text="here", text_with_ws="here",),
                    ],
                ),
            ],
        )

    def test_mark(self):
        """Test SSML mark tag"""
        processor = TextProcessor(default_lang="en_US", keep_whitespace=False)
        graph, root = processor(
            """
        <speak>
          <mark name="a"/>
          <p>
            <mark name="b" />
            <s>
              <mark name="c" />
              Mark <mark name="d" /> here
            </s>
            <mark name="e" />
          </p>
          <mark name="f" />
        </speak>
        """,
            ssml=True,
        )
        sentences = list(processor.sentences(graph, root, **WORDS_KWARGS))

        # Mark names are attached to appropriate elements
        self.assertEqual(
            sentences,
            [
                Sentence(
                    idx=0,
                    text="Mark here",
                    text_with_ws="Mark here",
                    text_spoken="Mark here",
                    marks_before=["a", "b"],
                    marks_after=["e", "f"],
                    words=[
                        Word(
                            idx=0,
                            sent_idx=0,
                            text="Mark",
                            text_with_ws="Mark",
                            marks_before=["c"],
                            marks_after=["d"],
                        ),
                        Word(idx=1, sent_idx=0, text="here", text_with_ws="here",),
                    ],
                ),
            ],
        )

    def test_missing_speak(self):
        """Test SSML with missing <speak> tag"""
        processor = TextProcessor()
        graph, root = processor("<s>hello</s><s>world</s>", ssml=True,)
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # <speak> is automatically added when XML fails to parse
        self.assertEqual(
            words,
            [
                Word(idx=0, sent_idx=0, text="hello", text_with_ws="hello",),
                Word(idx=0, sent_idx=1, text="world", text_with_ws="world",),
            ],
        )

    def test_adjacent_voice(self):
        """Test SSML with adjacent <voice> tags"""
        processor = TextProcessor()
        graph, root = processor(
            '<voice name="a">hello.</voice><voice name="b">world.</voice>', ssml=True,
        )
        words = list(processor.words(graph, root, major_breaks=False, **WORDS_KWARGS))

        self.assertEqual(
            words,
            [
                Word(idx=0, sent_idx=0, voice="a", text="hello", text_with_ws="hello",),
                Word(idx=0, sent_idx=1, voice="b", text="world", text_with_ws="world",),
            ],
        )

    def test_multiple_passes(self):
        """Test sentence that needs multiple passes to fully resolve"""
        processor = TextProcessor()
        graph, root = processor("ABCD-10")
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # 1) ABCD-10 -> ABCD 10
        # 2) ABCD 10 -> A B C D ten
        self.assertEqual(
            words,
            [
                Word(idx=0, text="A", text_with_ws="A ",),
                Word(idx=1, text="B", text_with_ws="B ",),
                Word(idx=2, text="C", text_with_ws="C ",),
                Word(idx=3, text="D", text_with_ws="D ",),
                Word(idx=4, text="ten", text_with_ws="ten",),
            ],
        )

    def test_number_nonfinite(self):
        """Test sentence with nan or inf"""
        processor = TextProcessor()
        graph, root = processor("nan inf")
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # Words should not be parsed as numbers
        self.assertEqual(
            words,
            [
                Word(idx=0, text="nan", text_with_ws="nan ",),
                Word(idx=1, text="inf", text_with_ws="inf",),
            ],
        )

    def test_override_initialism(self):
        """Test use of inline lexicon pronunciation to override an initialism"""
        processor = TextProcessor()
        graph, root = processor("ROOFUS")
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # Word is interpreted as initialism
        self.assertEqual(
            words,
            [
                Word(idx=0, text="R", text_with_ws="R ",),
                Word(idx=1, text="O", text_with_ws="O ",),
                Word(idx=2, text="O", text_with_ws="O ",),
                Word(idx=3, text="F", text_with_ws="F ",),
                Word(idx=4, text="U", text_with_ws="U ",),
                Word(idx=5, text="S", text_with_ws="S",),
            ],
        )

        graph, root = processor(
            """
        <speak>
          <lexicon>
            <lexeme>
              <grapheme>ROOFUS</grapheme>
              <phoneme>ɹ ˈu f ə s</phoneme>
            </lexeme>
          </lexicon>
          <s>ROOFUS</s>
        </speak>""",
            ssml=True,
        )
        words = list(processor.words(graph, root, **WORDS_KWARGS))

        # Word is *not* interpreted as initialism
        self.assertEqual(
            words, [Word(idx=0, text="ROOFUS", text_with_ws="ROOFUS",)],
        )


def print_graph_stderr(graph, root):
    """Print graph to stderr"""
    print_graph(graph, root, print_func=lambda *p: print(*p, file=sys.stderr))


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""Tests for SSML"""
import sys
import unittest

from gruut import sentences
from gruut.resources import _DIR
from gruut.utils import print_graph


class SSMLTestCase(unittest.TestCase):
    """Test cases for SSML"""

    def test_wikipedia_example(self):
        """Test SSML example from Wikipedia"""
        text = """<?xml version="1.0"?>
<speak xmlns="http://www.w3.org/2001/10/synthesis"
       xmlns:dc="http://purl.org/dc/elements/1.1/"
       version="1.0">
  <metadata>
    <dc:title xml:lang="en">Telephone Menu: Level 1</dc:title>
  </metadata>

  <p>
    <s xml:lang="en-US">
      <voice name="David" gender="male" age="25">
        For English, press <emphasis>one</emphasis>.
      </voice>
    </s>
    <s xml:lang="es-MX">
      <voice name="Miguel" gender="male" age="25">
        Para español, oprima el <emphasis>dos</emphasis>.
      </voice>
    </s>
  </p>

</speak>"""

        results = [
            (w.sent_idx, w.idx, w.lang, w.voice, w.text)
            for sent in sentences(text, ssml=True)
            for w in sent
        ]

        self.assertEqual(
            results,
            [
                (0, 0, "en-US", "David", "For"),
                (0, 1, "en-US", "David", "English"),
                (0, 2, "en-US", "David", ","),
                (0, 3, "en-US", "David", "press"),
                (0, 4, "en-US", "David", "one"),
                (0, 5, "en-US", "David", "."),
                (1, 0, "es-MX", "Miguel", "Para"),
                (1, 1, "es-MX", "Miguel", "español"),
                (1, 2, "es-MX", "Miguel", ","),
                (1, 3, "es-MX", "Miguel", "oprima"),
                (1, 4, "es-MX", "Miguel", "el"),
                (1, 5, "es-MX", "Miguel", "dos"),
                (1, 6, "es-MX", "Miguel", "."),
            ],
        )

    def test_lang_s(self):
        """Test lang on <s>"""
        text = """<?xml version="1.0" encoding="ISO-8859-1"?>
    <speak version="1.1" xmlns="http://www.w3.org/2001/10/synthesis"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
        xsi:schemaLocation="http://www.w3.org/2001/10/synthesis
                    http://www.w3.org/TR/speech-synthesis11/synthesis.xsd"
        xml:lang="en-US">
    <s>Today, 2/1/2000.</s>
    <!-- Today, February first two thousand -->
    <s xml:lang="it">Un mese fà, 2/1/2000.</s>
    <!-- Un mese fà, il due gennaio duemila -->
    <!-- One month ago, the second of January two thousand -->
    </speak>"""

        results = [
            (w.sent_idx, w.idx, w.lang, w.text)
            for sent in sentences(text, ssml=True)
            for w in sent
        ]

        self.assertEqual(
            results,
            [
                (0, 0, "en-US", "Today"),
                (0, 1, "en-US", ","),
                (0, 2, "en-US", "February"),
                (0, 3, "en-US", "first"),
                (0, 4, "en-US", ","),
                (0, 5, "en-US", "two"),
                (0, 6, "en-US", "thousand"),
                (0, 7, "en-US", "."),
                (1, 0, "it", "Un"),
                (1, 1, "it", "mese"),
                (1, 2, "it", "fà"),
                (1, 3, "it", ","),
                # no "il"
                (1, 4, "it", "due"),
                (1, 5, "it", "gennaio"),
                (1, 6, "it", "duemila"),
                (1, 7, "it", "."),
            ],
        )

    def test_phoneme(self):
        """Test manual phoneme insertion"""
        text = """<?xml version="1.0"?>
<speak version="1.1" xmlns="http://www.w3.org/2001/10/synthesis"
       xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
       xsi:schemaLocation="http://www.w3.org/2001/10/synthesis
                 http://www.w3.org/TR/speech-synthesis11/synthesis.xsd"
       xml:lang="en-US">
  <phoneme alphabet="ipa" ph="t&#x259;mei&#x325;&#x27E;ou&#x325;"> tomato </phoneme>
  <!-- This is an example of IPA using character entities -->
  <!-- Because many platform/browser/text editor combinations do not
       correctly cut and paste Unicode text, this example uses the entity
       escape versions of the IPA characters.  Normally, one would directly
       use the UTF-8 representation of these symbols: "təmei̥ɾou̥". -->
</speak>"""

        results = [
            (w.sent_idx, w.idx, w.lang, w.text, w.phonemes)
            for sent in sentences(text, ssml=True)
            for w in sent
        ]

        self.assertEqual(
            results,
            [(0, 0, "en-US", "tomato", ["t", "ə", "m", "e", "i̥", "ɾ", "o", "u̥"])],
        )

    def test_sentences(self):
        """Test <s>"""
        text = """<?xml version="1.0"?>
<speak version="1.1" xmlns="http://www.w3.org/2001/10/synthesis"
       xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
       xsi:schemaLocation="http://www.w3.org/2001/10/synthesis
                 http://www.w3.org/TR/speech-synthesis11/synthesis.xsd"
       xml:lang="en-US">
  <p>
    <s>This is the first sentence of the paragraph.</s>
    <s>Here's another sentence.</s>
  </p>
</speak>"""

        results = [
            (w.sent_idx, w.idx, w.text)
            for sent in sentences(text, ssml=True)
            for w in sent
        ]

        self.assertEqual(
            results,
            [
                (0, 0, "This"),
                (0, 1, "is"),
                (0, 2, "the"),
                (0, 3, "first"),
                (0, 4, "sentence"),
                (0, 5, "of"),
                (0, 6, "the"),
                (0, 7, "paragraph"),
                (0, 8, "."),
                (1, 0, "Here's"),
                (1, 1, "another"),
                (1, 2, "sentence"),
                (1, 3, "."),
            ],
        )

    def test_token(self):
        """Test explicit tokenization"""

        # NOTE: Added full stops
        text = """<?xml version="1.0"?>
<speak version="1.1" xmlns="http://www.w3.org/2001/10/synthesis"
       xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
       xsi:schemaLocation="http://www.w3.org/2001/10/synthesis
                 http://www.w3.org/TR/speech-synthesis11/synthesis.xsd"
       xml:lang="zh-CN">

  <!-- The Nanjing Changjiang River Bridge -->
  <token>南京市</token><token>长江大桥</token>。
  <!-- The mayor of Nanjing city, Jiang Daqiao -->
  南京市长<w>江大桥</w>。
  <!-- Shanghai is a metropolis -->
  上海是个<w>大都会</w>。
  <!-- Most Shanghainese will say something like that -->
  上海人<w>大都</w>会那么说。
</speak>"""

        results = [
            (w.sent_idx, w.idx, w.text)
            for sent in sentences(text, ssml=True)
            for w in sent
        ]

        self.assertEqual(
            results,
            [
                (0, 0, "南京市"),
                (0, 1, "长江大桥"),
                (0, 2, "。"),
                (1, 0, "南"),
                (1, 1, "京"),
                (1, 2, "市"),
                (1, 3, "长"),
                (1, 4, "江大桥"),
                (1, 5, "。"),
                (2, 0, "上"),
                (2, 1, "海"),
                (2, 2, "是"),
                (2, 3, "个"),
                (2, 4, "大都会"),
                (2, 5, "。"),
                (3, 0, "上"),
                (3, 1, "海"),
                (3, 2, "人"),
                (3, 3, "大都"),
                (3, 4, "会"),
                (3, 5, "那"),
                (3, 6, "么"),
                (3, 7, "说"),
                (3, 8, "。"),
            ],
        )

    def test_sub(self):
        """Test <sub>"""
        text = """<?xml version="1.0"?>
<speak version="1.1" xmlns="http://www.w3.org/2001/10/synthesis"
       xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
       xsi:schemaLocation="http://www.w3.org/2001/10/synthesis
                 http://www.w3.org/TR/speech-synthesis11/synthesis.xsd"
       xml:lang="en-US">
  <sub alias="World Wide Web Consortium">W3C</sub> is an international community
  <!-- World Wide Web Consortium -->
</speak>"""

        results = [
            (w.sent_idx, w.idx, w.text)
            for sent in sentences(text, ssml=True)
            for w in sent
        ]

        self.assertEqual(
            results,
            [
                (0, 0, "World"),
                (0, 1, "Wide"),
                (0, 2, "Web"),
                (0, 3, "Consortium"),
                (0, 4, "is"),
                (0, 5, "an"),
                (0, 6, "international"),
                (0, 7, "community"),
            ],
        )

    def test_lang_element(self):
        """Test <lang>"""
        text = """<?xml version="1.0"?>
<speak version="1.1" xmlns="http://www.w3.org/2001/10/synthesis"
       xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
       xsi:schemaLocation="http://www.w3.org/2001/10/synthesis
                 http://www.w3.org/TR/speech-synthesis11/synthesis.xsd"
       xml:lang="en-US">
  The French word for cat is <w xml:lang="fr">chat</w>.
  He prefers to eat pasta that is <lang xml:lang="it">al dente</lang>.
</speak>"""

        results = [
            (w.sent_idx, w.idx, w.lang, w.text)
            for sent in sentences(text, ssml=True)
            for w in sent
        ]

        self.assertEqual(
            results,
            [
                (0, 0, "en-US", "The"),
                (0, 1, "en-US", "French"),
                (0, 2, "en-US", "word"),
                (0, 3, "en-US", "for"),
                (0, 4, "en-US", "cat"),
                (0, 5, "en-US", "is"),
                (0, 6, "fr", "chat"),
                (0, 7, "en-US", "."),
                (1, 0, "en-US", "He"),
                (1, 1, "en-US", "prefers"),
                (1, 2, "en-US", "to"),
                (1, 3, "en-US", "eat"),
                (1, 4, "en-US", "pasta"),
                (1, 5, "en-US", "that"),
                (1, 6, "en-US", "is"),
                (1, 7, "it", "al"),
                (1, 8, "it", "dente"),
                (1, 9, "en-US", "."),
            ],
        )

    def test_lexicon(self):
        """Test <lexicon> and <lookup>"""
        text = """<?xml version="1.0"?>
<speak version="1.1"
       xmlns="http://www.w3.org/2001/10/synthesis"
       xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
       xsi:schemaLocation="http://www.w3.org/2001/10/synthesis
                 http://www.w3.org/TR/speech-synthesis11/synthesis.xsd"
       xml:lang="en-US">

  <lexicon xml:id="test" alphabet="ipa">
    <lexeme>
      <grapheme>
        tomato
      </grapheme>
      <phoneme>
        <!-- Individual phonemes are separated by whitespace -->
        t ə m ˈɑ t oʊ
      </phoneme>
    </lexeme>
    <lexeme role="fake-role">
      <grapheme>
        tomato
      </grapheme>
      <phoneme>
        <!-- Made up pronunciation for fake word role -->
        t ə m ˈi t oʊ
      </phoneme>
    </lexeme>
  </lexicon>

  <w>tomato</w>
  <lookup ref="test">
    <w>tomato</w>
    <w role="fake-role">tomato</w>
  </lookup>
</speak>"""

        results = [
            (w.sent_idx, w.idx, w.phonemes)
            for sent in sentences(text, ssml=True)
            for w in sent
        ]

        self.assertEqual(
            results,
            [
                (0, 0, ["t", "ə", "m", "ˈeɪ", "t", "oʊ"]),
                (0, 1, ["t", "ə", "m", "ˈɑ", "t", "oʊ"]),
                (0, 2, ["t", "ə", "m", "ˈi", "t", "oʊ"]),
            ],
        )

    def test_lexicon_external(self):
        """Test <lexicon> from URI"""
        lexicon_path = (_DIR.parent / "etc" / "sample_lexicon.xml").absolute()

        text = f"""<?xml version="1.0"?>
<speak version="1.1"
       xmlns="http://www.w3.org/2001/10/synthesis"
       xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
       xsi:schemaLocation="http://www.w3.org/2001/10/synthesis
                 http://www.w3.org/TR/speech-synthesis11/synthesis.xsd"
       xml:lang="en-US">

  <lexicon xml:id="test" alphabet="ipa" uri="file://{lexicon_path}" />

  <lookup ref="test">
    <w>tomato</w>
  </lookup>
</speak>"""

        results = [
            (w.sent_idx, w.idx, w.phonemes)
            for sent in sentences(text, ssml=True)
            for w in sent
        ]

        self.assertEqual(
            results, [(0, 0, ["t", "ə", "m", "e", "i̥", "ɾ", "o", "u̥"])],
        )


def print_graph_stderr(graph, root):
    """Print graph to stderr"""
    print_graph(graph, root, print_func=lambda *p: print(*p, file=sys.stderr))


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""Converts a Phonetisaurus G2P corpus to an sqlite database"""
import argparse
import sqlite3

# -----------------------------------------------------------------------------


def main():
    """Main entry point"""
    parser = argparse.ArgumentParser(prog="corpus2db.py")
    parser.add_argument(
        "--corpus", required=True, help="Phonetisaurus G2P corpus file to read"
    )
    parser.add_argument("--database", required=True, help="SQLite database to write")
    parser.add_argument(
        "--no-drop",
        action="store_true",
        help="Don't drop existing g2p_alignments table",
    )
    args = parser.parse_args()

    # -------------------------------------------------------------------------

    conn = sqlite3.connect(args.database)

    g2p_alignments = {}
    with open(args.corpus, "r", encoding="utf-8") as corpus_file:
        for line in corpus_file:
            line = line.strip()
            if not line:
                continue

            word = ""

            # Parse line
            parts = line.split()
            for part in parts:
                # Assume default delimiters:
                # } separates input/output
                # | separates input/output tokens
                # _ indicates empty output
                part_in, _part_out = part.split("}")
                part_ins = part_in.split("|")
                word += "".join(part_ins)

            if word and (word not in g2p_alignments):
                g2p_alignments[word] = line

    # Add to database
    with conn:
        if not args.no_drop:
            conn.execute("DROP TABLE IF EXISTS g2p_alignments")

        conn.execute(
            "CREATE TABLE IF NOT EXISTS g2p_alignments "
            + "(id INTEGER PRIMARY KEY AUTOINCREMENT, word TEXT, alignment TEXT);"
        )

        for word, alignment in g2p_alignments.items():
            conn.execute(
                "INSERT INTO g2p_alignments (word, alignment) VALUES (?, ?)",
                (word, alignment),
            )

    print("Added", len(g2p_alignments), "alignments to", args.database)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Language-specific settings"""
import logging
import re
import sqlite3
import typing
from pathlib import Path

import networkx as nx

from gruut.const import PHONEMES_TYPE, GraphType, SentenceNode, Time
from gruut.g2p import GraphemesToPhonemes
from gruut.phonemize import SqlitePhonemizer
from gruut.pos import PartOfSpeechTagger
from gruut.text_processor import InterpretAsFormat, TextProcessorSettings
from gruut.utils import find_lang_dir, remove_non_word_chars, resolve_lang

_LOGGER = logging.getLogger("gruut")

# -----------------------------------------------------------------------------


def get_settings(
    lang: str,
    search_dirs: typing.Optional[typing.Iterable[typing.Union[str, Path]]] = None,
    lang_dir: typing.Optional[typing.Union[str, Path]] = None,
    model_prefix: typing.Optional[str] = None,
    load_pos_tagger: bool = True,
    load_phoneme_lexicon: bool = True,
    load_g2p_guesser: bool = True,
    **settings_args,
) -> TextProcessorSettings:
    """Get settings for a specific language"""
    model_prefix = model_prefix or ""

    # Resolve language
    if model_prefix:
        # espeak
        lang_model_prefix = model_prefix
        lang_only = lang
    elif "/" in lang:
        # en-us/espeak
        lang_only, lang_model_prefix = lang.split("/", maxsplit=1)
    else:
        # en-us
        lang_only = lang
        lang_model_prefix = ""

    # en_US -> en-us
    lang_only = resolve_lang(lang_only)

    if lang_dir is None:
        # Search for language data files
        lang_dir = find_lang_dir(lang_only, search_dirs=search_dirs)

    if lang_dir is not None:
        lang_dir = Path(lang_dir)

        # Part of speech tagger
        if load_pos_tagger and ("get_parts_of_speech" not in settings_args):
            pos_model_path = lang_dir / "pos" / "model.crf"
            if pos_model_path.is_file():
                # POS tagger model will load on first use
                settings_args["get_parts_of_speech"] = DelayedPartOfSpeechTagger(
                    pos_model_path
                )
            else:
                _LOGGER.debug(
                    "(%s) no part of speech tagger found at %s", lang, pos_model_path,
                )

        # Phonemizer
        if load_phoneme_lexicon and ("lookup_phonemes" not in settings_args):
            lexicon_db_path = lang_dir / lang_model_prefix / "lexicon.db"
            if lexicon_db_path.is_file():
                # Transformations to apply to words when they can't be found in the lexicon
                phonemizer_args = {
                    "word_transform_funcs": [
                        remove_non_word_chars,
                        lambda s: remove_non_word_chars(s.lower()),
                    ],
                    "casing_func": str.lower,
                }

                settings_args["lookup_phonemes"] = DelayedSqlitePhonemizer(
                    lexicon_db_path, **phonemizer_args
                )
            else:
                _LOGGER.debug(
                    "(%s) no phoneme lexicon database found at %s",
                    lang,
                    lexicon_db_path,
                )

        # Grapheme to phoneme model
        if load_g2p_guesser and ("guess_phonemes" not in settings_args):
            g2p_model_path = lang_dir / lang_model_prefix / "g2p" / "model.crf"
            if g2p_model_path.is_file():
                settings_args["guess_phonemes"] = DelayedGraphemesToPhonemes(
                    g2p_model_path, transform_func=str.lower
                )

            else:
                _LOGGER.debug(
                    "(%s) no grapheme to phoneme CRF model found at %s",
                    lang,
                    g2p_model_path,
                )

    # ---------------------------------
    # Create language-specific settings
    # ---------------------------------

    if lang_only == "ar":
        # Arabic
        return get_ar_settings(lang_dir, **settings_args)

    if lang_only == "cs-cz":
        # Czech
        return get_cs_settings(lang_dir, **settings_args)

    if lang_only in {"en-us", "en-gb"}:
        # English
        return get_en_us_settings(lang_dir, **settings_args)

    if lang_only == "de-de":
        # German
        return get_de_settings(lang_dir, **settings_args)

    if lang_only in {"es-es", "es-mx"}:
        # Spanish
        return get_es_settings(lang_dir, **settings_args)

    if lang_only == "fa":
        # Farsi
        return get_fa_settings(lang_dir, **settings_args)

    if lang_only == "fr-fr":
        # French
        return get_fr_settings(lang_dir, **settings_args)

    if lang_only == "it-it":
        # Italian
        return get_it_settings(lang_dir, **settings_args)

    if lang_only == "lb":
        # Lëtzebuergesch
        return get_lb_settings(lang_dir, **settings_args)

    if lang_only == "nl":
        # Dutch
        return get_nl_settings(lang_dir, **settings_args)

    if lang_only == "pt":
        # Portuguese
        return get_pt_settings(lang_dir, **settings_args)

    if lang_only == "ru-ru":
        # Russian
        return get_ru_settings(lang_dir, **settings_args)

    if lang_only == "sv-se":
        # Swedish
        return get_sv_settings(lang_dir, **settings_args)

    if lang_only == "sw":
        # Swahili
        return get_sw_settings(lang_dir, **settings_args)

    if lang_only == "zh-cn":
        # Chinese
        return get_zh_settings(lang_dir, **settings_args)

    # Default settings only
    return TextProcessorSettings(lang=lang, **settings_args)


# -----------------------------------------------------------------------------
# Arabic (ar, اَلْعَرَبِيَّةُ)
# -----------------------------------------------------------------------------


class ArabicPreProcessText:
    """Pre-processes text using mishkal"""

    def __call__(self, text: str) -> str:
        try:
            import mishkal.tashkeel

            # Load vocalizer
            if not hasattr(self, "vocalizer"):
                vocalizer = mishkal.tashkeel.TashkeelClass()
                setattr(self, "vocalizer", vocalizer)
            else:
                vocalizer = getattr(self, "vocalizer")

            assert vocalizer is not None

            # Add diacritics
            text = vocalizer.tashkeel(text)
        except ImportError:
            _LOGGER.warning("mishkal is highly recommended for language 'ar'")
            _LOGGER.warning("pip install 'mishkal>=0.4.0'")

        return text


def get_ar_settings(lang_dir=None, **settings_args) -> TextProcessorSettings:
    """Create settings for Arabic"""
    settings_args = {
        "major_breaks": {".", "؟", "!"},
        "minor_breaks": {"،", ";", ":"},
        "word_breaks": {"-", "_"},
        "begin_punctuations": {'"', "“", "«", "[", "(", "<", "„"},
        "end_punctuations": {'"', "”", "»", "]", ")", ">"},
        "default_date_format": InterpretAsFormat.DATE_DMY,
        "replacements": [("’", "'")],  # normalize apostrophe
        "pre_process_text": ArabicPreProcessText(),
        **settings_args,
    }
    return TextProcessorSettings(lang="ar", **settings_args)


# -----------------------------------------------------------------------------
# Czech (cs-cz, čeština)
# -----------------------------------------------------------------------------


def get_cs_settings(lang_dir=None, **settings_args) -> TextProcessorSettings:
    """Create settings for Czech"""
    settings_args = {
        "major_breaks": {".", "?", "!"},
        "minor_breaks": {",", ";", ":"},
        "word_breaks": {"-", "_"},
        "begin_punctuations": {'"', "“", "«", "[", "(", "<", "’", "„"},
        "end_punctuations": {'"', "”", "»", "]", ")", ">", "’"},
        "default_currency": "EUR",
        "default_date_format": InterpretAsFormat.DATE_DMY,
        "replacements": [("’", "'")],  # normalize apostrophe
        **settings_args,
    }
    return TextProcessorSettings(lang="cs_CZ", **settings_args)


# -----------------------------------------------------------------------------
# English (en-us, en-gb)
# -----------------------------------------------------------------------------


# TTS and T.T.S.
EN_INITIALISM_PATTERN = re.compile(r"^\s*[A-Z]{2,}\s*$")
EN_INITIALISM_DOTS_PATTERN = re.compile(r"^(?:\s*[a-zA-Z]\.){1,}\s*$")

EN_NON_WORD_PATTERN = re.compile(r"^(\W|_)+$")
EN_ORDINAL_PATTERN = re.compile(r"^(-?[0-9][0-9,]*)(?:st|nd|rd|th).*$")

EN_TIME_PATTERN = re.compile(
    r"""^((0?[0-9])|(1[0-1])|(1[2-9])|(2[0-3]))  # hours
         (?::
         ([0-5][0-9]))?                          # minutes
         \s*(a\\.m\\.|am|pm|p\\.m\\.|a\\.m|p\\.m)? # am/pm
         $""",
    re.IGNORECASE | re.X,
)


def en_is_initialism(text: str) -> bool:
    """True if text is of the form TTS or T.T.S."""
    return (EN_INITIALISM_PATTERN.match(text) is not None) or (
        EN_INITIALISM_DOTS_PATTERN.match(text) is not None
    )


def en_get_ordinal(text: str) -> typing.Optional[int]:
    """Parse English ordinal string (e.g., 1st -> 1)"""
    match = EN_ORDINAL_PATTERN.match(text)
    if match is not None:
        return int(re.sub(r"[^0-9]", "", match.group(1)))

    return None


def en_parse_time(text: str) -> typing.Optional[Time]:
    """Parse English clock time (e.g. 4:01pm)"""
    match = EN_TIME_PATTERN.match(text.strip().lower())
    if match is None:
        return None

    hours = int(match.group(1))
    maybe_minutes = match.group(6)
    minutes = 0 if maybe_minutes is None else int(maybe_minutes)
    period = match.group(7)

    if period is not None:
        # Normalize period
        if "a" in period:
            period = "A.M."
        else:
            period = "P.M."
    else:
        if ":" not in text:
            # Require a colon if no period is specified to avoid parsing plain
            # numbers like "1" into time expressions.
            return None

    return Time(hours=hours, minutes=minutes, period=period)


def en_verbalize_time(time: Time) -> typing.Iterable[str]:
    """Convert time into words"""

    hour = time.hours
    past_noon = hour >= 12

    if hour > 12:
        hour -= 12
    elif hour == 0:
        hour = 12
        past_noon = True

    yield str(hour)

    minute = time.minutes
    if minute > 0:
        if minute < 10:
            yield "oh"

        yield str(minute)

    if time.period is None:
        if past_noon:
            yield "P.M."
        else:
            yield "A.M."
    else:
        yield time.period


def get_en_us_settings(lang_dir=None, **settings_args) -> TextProcessorSettings:
    """Create settings for English"""
    settings_args = {
        "major_breaks": {".", "?", "!"},
        "minor_breaks": {",", ";", ":", "..."},
        "word_breaks": {"-", "_"},
        "begin_punctuations": {'"', "'", "“", "«", "[", "(", "<", "*", "_"},
        "end_punctuations": {'"', "'", "”", "»", "]", ")", ">", "*", "_"},
        "default_currency": "USD",
        "default_date_format": "{m} {o}, {y}",
        "is_initialism": en_is_initialism,
        "split_initialism": lambda text: list(text.replace(".", "")),
        "is_non_word": lambda text: EN_NON_WORD_PATTERN.match(text) is not None,
        "get_ordinal": en_get_ordinal,
        "parse_time": en_parse_time,
        "verbalize_time": en_verbalize_time,
        "replacements": [("’", "'")],  # normalize apostrophe
        "abbreviations": {
            r"^([cC])o\.": r"\1ompany",  # co. -> company
            r"^([dD])r\.": r"\1octor",  # dr. -> doctor
            r"^([dD])rs\.": r"\1octors",  # drs. -> doctors
            r"^([jJ])r\.('s)?": r"\1unior\2",  # jr. -> junior
            r"^([lL])td\.": r"\1imited",  # -> ltd. -> limited
            r"^([mM])r\.": r"\1ister",  # -> mr. -> mister
            r"^([mM])s\.": r"\1iss",  # -> ms. -> miss
            r"^([mM])rs\.": r"\1issus",  # -> mrs. -> missus
            r"^([sS])t\.": r"\1treet",  # -> st. -> street
            r"^([vV])s\.?": r"\1ersus",  # -> vs. -> versus
            r"(.*\d)%": r"\1 percent",  # % -> percent
            r"^&(\s*)$": r"and\1",  # &-> and
            r"^([mM])t\.": r"\1ount",  # -> mt. -> mount
        },
        "spell_out_words": {
            ".": "dot",
            "-": "dash",
            "@": "at",
            "*": "star",
            "+": "plus",
            "/": "slash",
        },
        **settings_args,
    }

    return TextProcessorSettings(lang="en_US", **settings_args)


# -----------------------------------------------------------------------------
# German (de-de)
# -----------------------------------------------------------------------------


def get_de_settings(lang_dir=None, **settings_args) -> TextProcessorSettings:
    """Create settings for German"""
    settings_args = {
        "major_breaks": {".", "?", "!"},
        "minor_breaks": {",", ";", ":", "..."},
        "word_breaks": {"-", "_"},
        "begin_punctuations": {'"', "“", "«", "[", "(", "<", "’", "„"},
        "end_punctuations": {'"', "”", "»", "]", ")", ">", "’"},
        "default_currency": "EUR",
        "default_date_format": InterpretAsFormat.DATE_DMY_ORDINAL,
        "replacements": [("’", "'")],  # normalize apostrophe
        **settings_args,
    }
    return TextProcessorSettings(lang="de_DE", **settings_args)


# -----------------------------------------------------------------------------
# Spanish (es-es, Español)
# -----------------------------------------------------------------------------


def get_es_settings(lang_dir=None, **settings_args) -> TextProcessorSettings:
    """Create settings for Spanish"""
    settings_args = {
        "major_breaks": {".", "?", "!"},
        "minor_breaks": {",", ";", ":", "..."},
        "word_breaks": {"-", "_"},
        "begin_punctuations": {'"', "“", "«", "[", "(", "<", "¡", "¿"},
        "end_punctuations": {'"', "”", "»", "]", ")", ">"},
        "default_currency": "EUR",
        "default_date_format": InterpretAsFormat.DATE_DMY,
        "replacements": [("’", "'")],  # normalize apostrophe
        **settings_args,
    }
    return TextProcessorSettings(lang="es_ES", **settings_args)


# -----------------------------------------------------------------------------
# Farsi/Persian (fa, فارسی)
# -----------------------------------------------------------------------------


class FarsiPartOfSpeechTagger:
    """Add POS tags with hazm"""

    def __init__(self, lang_dir: Path):
        self.lang_dir = lang_dir

    def __call__(self, words: typing.Sequence[str]) -> typing.Sequence[str]:
        pos_tags = []

        try:
            import hazm

            # Load normalizer
            normalizer = getattr(self, "normalizer", None)
            if normalizer is None:
                normalizer = hazm.Normalizer()
                setattr(self, "normalizer", normalizer)

            # Load tagger
            tagger = getattr(self, "tagger", None)
            if tagger is None:
                # Load part of speech tagger
                model_path = self.lang_dir / "pos" / "postagger.model"
                tagger = hazm.POSTagger(model=str(model_path))
                setattr(self, "tagger", tagger)

            text = " ".join(words)
            for sentence in hazm.sent_tokenize(normalizer.normalize(text)):
                for _word, pos in tagger.tag(hazm.word_tokenize(sentence)):
                    pos_tags.append(pos)
        except ImportError:
            _LOGGER.warning("hazm is highly recommended for language 'fa'")
            _LOGGER.warning("pip install 'hazm>=0.7.0'")

        return pos_tags


def fa_post_process_sentence(
    graph: GraphType, sent_node: SentenceNode, settings: TextProcessorSettings
):
    """Add e̞ for genitive case"""
    from gruut.text_processor import DATA_PROP, WordNode

    for dfs_node in nx.dfs_preorder_nodes(graph, sent_node.node):
        if not graph.out_degree(dfs_node) == 0:
            # Only leave
            continue

        node = graph.nodes[dfs_node][DATA_PROP]
        if isinstance(node, WordNode):
            word = typing.cast(WordNode, node)
            if word.phonemes and (word.pos == "Ne"):
                if isinstance(word.phonemes, list):
                    word.phonemes.append("e̞")
                else:
                    word.phonemes = list(word.phonemes) + ["e̞"]


def get_fa_settings(lang_dir=None, **settings_args) -> TextProcessorSettings:
    """Create settings for Farsi"""
    settings_args = {
        "major_breaks": {".", "؟", "!"},
        "minor_breaks": {",", ";", ":"},
        "word_breaks": {"-", "_"},
        "begin_punctuations": {'"', "“", "«", "[", "(", "<", "’", "„"},
        "end_punctuations": {'"', "”", "»", "]", ")", ">", "’"},
        "default_date_format": InterpretAsFormat.DATE_DMY,
        "replacements": [("’", "'")],  # normalize apostrophe
        "post_process_sentence": fa_post_process_sentence,
        **settings_args,
    }

    if (lang_dir is not None) and ("get_parts_of_speech" not in settings_args):
        settings_args["get_parts_of_speech"] = FarsiPartOfSpeechTagger(lang_dir)

    return TextProcessorSettings(lang="fa", **settings_args)


# -----------------------------------------------------------------------------
# French (fr-fr, Français)
# -----------------------------------------------------------------------------


def fr_post_process_sentence(
    graph: GraphType, sent_node: SentenceNode, settings: TextProcessorSettings
):
    """Add liasons to phonemes"""
    from gruut.text_processor import DATA_PROP, WordNode
    from gruut.utils import sliding_window

    words = []
    for dfs_node in nx.dfs_preorder_nodes(graph, sent_node.node):
        if not graph.out_degree(dfs_node) == 0:
            # Only leave
            continue

        node = graph.nodes[dfs_node][DATA_PROP]
        if isinstance(node, WordNode):
            word_node = typing.cast(WordNode, node)
            words.append(word_node)

    for word1, word2 in sliding_window(words, 2):
        if word2 is None:
            continue

        if not (word1.text and word1.phonemes and word2.text and word2.phonemes):
            continue

        liason = False

        # Conditions to meet for liason check:
        # 1) word 1 ends with a silent consonant
        # 2) word 2 starts with a vowel (phoneme)

        last_char1 = word1.text[-1]
        ends_silent_consonant = fr_has_silent_consonant(last_char1, word1.phonemes[-1])
        starts_vowel = fr_is_vowel(word2.phonemes[0])

        if ends_silent_consonant and starts_vowel:
            # Handle mandatory liason cases
            # https://www.commeunefrancaise.com/blog/la-liaison

            if word1.text == "et":
                # No liason
                pass
            elif word1.pos in {"DET", "NUM"}:
                # Determiner/adjective -> noun
                liason = True
            elif (word1.pos == "PRON") and (word2.pos in {"AUX", "VERB"}):
                # Pronoun -> verb
                liason = True
            elif (word1.pos == "ADP") or (word1.text == "très"):
                # Preposition
                liason = True
            elif (word1.pos == "ADJ") and (word2.pos in {"NOUN", "PROPN"}):
                # Adjective -> noun
                liason = True
            elif word1.pos in {"AUX", "VERB"}:
                # Verb -> vowel
                liason = True

        if liason:
            # Apply liason
            # s -> z
            # p -> p
            # d|t -> d
            liason_pron = word1.phonemes

            if last_char1 in {"s", "x", "z"}:
                liason_pron.append("z")
            elif last_char1 == "d":
                liason_pron.append("t")
            elif last_char1 in {"t", "p", "n"}:
                # Final phoneme is same as char
                liason_pron.append(last_char1)


def fr_has_silent_consonant(last_char: str, last_phoneme: str) -> bool:
    """True if last consonant is silent in French"""
    # Credit: https://github.com/Remiphilius/PoemesProfonds/blob/master/lecture.py

    if last_char in {"d", "p", "t"}:
        return last_phoneme != last_char
    if last_char == "r":
        return last_phoneme != "ʁ"
    if last_char in {"s", "x", "z"}:
        return last_phoneme not in {"s", "z"}
    if last_char == "n":
        return last_phoneme not in {"n", "ŋ"}

    return False


def fr_is_vowel(phoneme: str) -> bool:
    """True if phoneme is a French vowel"""
    return phoneme in {
        "i",
        "y",
        "u",
        "e",
        "ø",
        "o",
        "ə",
        "ɛ",
        "œ",
        "ɔ",
        "a",
        "ɔ̃",
        "ɛ̃",
        "ɑ̃",
        "œ̃",
    }


def get_fr_settings(lang_dir=None, **settings_args) -> TextProcessorSettings:
    """Create settings for French"""
    settings_args = {
        "major_breaks": {".", "?", "!"},
        "minor_breaks": {",", ";", ":", "..."},
        "word_breaks": {"-", "_"},
        "begin_punctuations": {'"', "“", "«", "[", "(", "<", "„"},
        "end_punctuations": {'"', "”", "»", "]", ")", ">"},
        "default_currency": "EUR",
        "default_date_format": InterpretAsFormat.DATE_DMY_ORDINAL,
        "replacements": [("’", "'")],  # normalize apostrophe
        "post_process_sentence": fr_post_process_sentence,
        **settings_args,
    }
    return TextProcessorSettings(lang="fr_FR", **settings_args)


# -----------------------------------------------------------------------------
# Italian (it-it, Italiano)
# -----------------------------------------------------------------------------


def get_it_settings(lang_dir=None, **settings_args) -> TextProcessorSettings:
    """Create settings for Italian"""
    settings_args = {
        "major_breaks": {".", "?", "!"},
        "minor_breaks": {",", ";", ":", "..."},
        "word_breaks": {"-", "_"},
        "begin_punctuations": {'"', "“", "«", "[", "(", "<", "„"},
        "end_punctuations": {'"', "”", "»", "]", ")", ">"},
        "default_currency": "EUR",
        "default_date_format": InterpretAsFormat.DATE_DMY,
        "replacements": [("’", "'")],  # normalize apostrophe
        "post_process_sentence": fr_post_process_sentence,
        **settings_args,
    }
    return TextProcessorSettings(lang="it_IT", **settings_args)


# -----------------------------------------------------------------------------
# Luxembourgish (lb, Lëtzebuergesch)
# -----------------------------------------------------------------------------


def get_lb_settings(lang_dir=None, **settings_args) -> TextProcessorSettings:
    """Create settings for Luxembourgish"""
    settings_args = {
        "major_breaks": {".", "?", "!"},
        "minor_breaks": {",", ";", ":", "..."},
        "word_breaks": {"-", "_"},
        "begin_punctuations": {'"', "“", "«", "[", "(", "<", "„"},
        "end_punctuations": {'"', "”", "»", "]", ")", ">"},
        "default_currency": "EUR",
        "default_date_format": InterpretAsFormat.DATE_DMY,
        "replacements": [("’", "'")],  # normalize apostrophe
        "babel_locale": "lb",
        **settings_args,
    }
    return TextProcessorSettings(lang="lb", **settings_args)


# -----------------------------------------------------------------------------
# Dutch (nl, Nederlands)
# -----------------------------------------------------------------------------


def get_nl_settings(lang_dir=None, **settings_args) -> TextProcessorSettings:
    """Create settings for Dutch"""
    settings_args = {
        "major_breaks": {".", "?", "!"},
        "minor_breaks": {",", ";", ":", "..."},
        "word_breaks": {"-", "_"},
        "begin_punctuations": {'"', "“", "«", "[", "(", "<", "„"},
        "end_punctuations": {'"', "”", "»", "]", ")", ">"},
        "default_currency": "EUR",
        "default_date_format": InterpretAsFormat.DATE_DMY,
        "replacements": [("’", "'")],  # normalize apostrophe
        **settings_args,
    }
    return TextProcessorSettings(lang="nl", **settings_args)


# -----------------------------------------------------------------------------
# Portuguese (pt, Português)
# -----------------------------------------------------------------------------


def get_pt_settings(lang_dir=None, **settings_args) -> TextProcessorSettings:
    """Create default settings for Portuguese"""
    settings_args = {
        "major_breaks": {".", "?", "!"},
        "minor_breaks": {",", ";", ":", "..."},
        "word_breaks": {"-", "_"},
        "begin_punctuations": {'"', "“", "«", "[", "(", "<", "„"},
        "end_punctuations": {'"', "”", "»", "]", ")", ">"},
        "default_currency": "EUR",
        "default_date_format": InterpretAsFormat.DATE_DMY,
        "replacements": [("’", "'")],  # normalize apostrophe
        **settings_args,
    }
    return TextProcessorSettings(lang="pt", **settings_args)


# -----------------------------------------------------------------------------
# Russian (ru, Русский)
# -----------------------------------------------------------------------------


def get_ru_settings(lang_dir=None, **settings_args) -> TextProcessorSettings:
    """Create settings for Russian"""
    settings_args = {
        "major_breaks": {".", "?", "!"},
        "minor_breaks": {",", ";", ":"},
        "word_breaks": {"-", "_"},
        "begin_punctuations": {'"', "“", "«", "[", "(", "<", "„"},
        "end_punctuations": {'"', "”", "»", "]", ")", ">"},
        "default_currency": "RUB",
        "default_date_format": InterpretAsFormat.DATE_DMY,
        "replacements": [("’", "'")],  # normalize apostrophe
        **settings_args,
    }
    return TextProcessorSettings(lang="ru_RU", **settings_args)


# -----------------------------------------------------------------------------
# Swedish (sv-se, svenska)
# -----------------------------------------------------------------------------


def get_sv_settings(lang_dir=None, **settings_args) -> TextProcessorSettings:
    """Create settings for Swedish"""
    settings_args = {
        "major_breaks": {".", "?", "!"},
        "minor_breaks": {",", ";", ":", "..."},
        "word_breaks": {"-", "_"},
        "begin_punctuations": {'"', "“", "«", "[", "(", "<", "„"},
        "end_punctuations": {'"', "”", "»", "]", ")", ">"},
        "default_date_format": InterpretAsFormat.DATE_DMY,
        "replacements": [("’", "'")],  # normalize apostrophe
        **settings_args,
    }
    return TextProcessorSettings(lang="sv_SE", **settings_args)


# -----------------------------------------------------------------------------
# Swahili (sw, Kiswahili)
# -----------------------------------------------------------------------------


def get_sw_settings(lang_dir=None, **settings_args) -> TextProcessorSettings:
    """Create settings for Swahili"""
    settings_args = {
        "major_breaks": {".", "?", "!"},
        "minor_breaks": {",", ";", ":"},
        "word_breaks": {"-", "_"},
        "begin_punctuations": {'"', "“", "«", "[", "(", "<", "„"},
        "end_punctuations": {'"', "”", "»", "]", ")", ">"},
        "default_date_format": InterpretAsFormat.DATE_DMY,
        "replacements": [("’", "'")],  # normalize apostrophe
        **settings_args,
    }
    return TextProcessorSettings(lang="sw", **settings_args)


# -----------------------------------------------------------------------------
# Chinese (zh-cn, 汉语)
# -----------------------------------------------------------------------------


def get_zh_settings(lang_dir=None, **settings_args) -> TextProcessorSettings:
    """Create settings for Chinese"""

    # https://en.wikipedia.org/wiki/Chinese_punctuation
    settings_args = {
        "major_breaks": {"。", "！", "？"},
        "minor_breaks": {"；", "：", "，", "、", "……"},
        "begin_punctuations": {"（", "［", "【", "「", "﹁", '"', "《", "〈"},
        "end_punctuations": {"）", "］", " 】", "」", "﹂", '"', "》", "〉"},
        "word_breaks": {"‧"},
        "split_words": list,
        "join_str": "",
        **settings_args,
    }
    return TextProcessorSettings(lang="zh_CN", **settings_args)


# -----------------------------------------------------------------------------


class DelayedGraphemesToPhonemes:
    """Grapheme to phoneme guesser that loads on first use"""

    def __init__(
        self,
        model_path: typing.Union[str, Path],
        transform_func: typing.Optional[typing.Callable[[str], str]] = None,
        **g2p_args,
    ):
        self.model_path = model_path
        self.g2p: typing.Optional[GraphemesToPhonemes] = None
        self.transform_func = transform_func
        self.g2p_args = g2p_args

    def __call__(
        self, word: str, role: typing.Optional[str] = None
    ) -> typing.Optional[PHONEMES_TYPE]:
        if self.g2p is None:
            _LOGGER.debug(
                "Loading grapheme to phoneme CRF model from %s", self.model_path
            )
            self.g2p = GraphemesToPhonemes(self.model_path, **self.g2p_args)

        assert self.g2p is not None

        if self.transform_func is not None:
            word = self.transform_func(word)

        return self.g2p(word)


class DelayedPartOfSpeechTagger:
    """POS tagger that loads on first use"""

    def __init__(self, model_path: typing.Union[str, Path], **tagger_args):

        self.model_path = Path(model_path)
        self.tagger: typing.Optional[PartOfSpeechTagger] = None
        self.tagger_args = tagger_args

    def __call__(self, words: typing.Sequence[str]) -> typing.Sequence[str]:
        if self.tagger is None:
            _LOGGER.debug("Loading part of speech tagger from %s", self.model_path)
            self.tagger = PartOfSpeechTagger(self.model_path, **self.tagger_args)

        assert self.tagger is not None
        return self.tagger(words)


class DelayedSqlitePhonemizer:
    """Phonemizer that loads on first use"""

    def __init__(self, db_path: typing.Union[str, Path], **phonemizer_args):

        self.db_path = Path(db_path)
        self.phonemizer: typing.Optional[SqlitePhonemizer] = None
        self.phonemizer_args = phonemizer_args

    def __call__(
        self, word: str, role: typing.Optional[str] = None, do_transforms: bool = True
    ) -> typing.Optional[PHONEMES_TYPE]:
        if self.phonemizer is None:
            _LOGGER.debug("Connecting to lexicon database at %s", self.db_path)
            db_conn = sqlite3.connect(str(self.db_path))
            self.phonemizer = SqlitePhonemizer(db_conn=db_conn, **self.phonemizer_args)

        assert self.phonemizer is not None
        return self.phonemizer(word, role=role, do_transforms=do_transforms)
